"""
Background substraction is a necssary component of reflectometry reduction, where the background scattering is removed from the reflected intensity.

Herein are some function to enable that for a two-dimensional detector image.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)
# pylint: disable=R0913,C0103

import numpy as np
from scipy.stats import multivariate_normal, norm
from scipy.optimize import curve_fit
from uncertainties import unumpy as unp


def bivariate_normal(data, mu_1, mu_2, sigma_1, sigma_2, offset, factor):
    """
    Produce a bivariate normal distribution.

    *Note*: the covariance of the two dimensions is assumed to be zero to unsure greater stability.

    Args:
        data (:py:attr:`array_like`): Two-dimensional abscissa data.
        mu_1 (:py:attr:`float`): Mean in dimension 0 (horizontal).
        mu_2 (:py:attr:`float`): Mean in dimension 1 (vertical).
        sigma_1 (:py:attr:`float`): Variance in dimension 0 (horizontal).
        sigma_2 (:py:attr:`float`): Variance in dimension 1 (vertical).
        offset (:py:attr:`float`): Offset from the 0 for the ordinate, this is the background level.
        factor (:py:attr:`float`): Multiplicative factor for area of normal distribution.

    Returns:
        :py:attr:`array_like`: Flattened ordinate data for bivariate normal distribution.
    """
    # Setting the data up in the correct format
    pos = np.empty(data[0].shape + (2,))
    pos[:, :, 0] = data[0]
    pos[:, :, 1] = data[1]
    # Creation of the bivariate normal distribution
    binormal = multivariate_normal([mu_1, mu_2], [[sigma_1, 0], [0, sigma_2]])
    return offset + binormal.pdf(pos).flatten() * factor


def fit_gaussian_2d(image, image_e, p0=None, bounds=None):
    """
    Fit a two-dimensional Gaussian function with some ordinate offset to an
    image with an uncertainty and return the results, index of offset and
    vertical distribution width.

    Args:
        image (:py:attr:`array_like`): The data to fit the Gaussian to.
        image_e (:py:attr:`array_like`): The data uncertainty.
        p0 (:py:attr:`list`, optional): An initial guess at the parameters. Defaults to values based on the image.
        bounds (:py:attr:`list` of :py:attr:`tuple`, optional): Bounds for the fitting. Defaults to values based on the image.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: The results (with uncertainties) for each of the 6 parameters fit.
            - :py:attr:`int`: The index of the offset.
            - :py:attr:`int`: The index of the vertical distribution width.
    """
    # Setting default values
    if p0 is None:
        p0 = [
            image.shape[0] / 2,
            image.shape[1] / 2,
            1,
            1,
            image.min(),
            image.max(),
        ]
    if bounds is None:
        bounds = (
            0,
            [image.shape[0], image.shape[1], 100, 100, image.max(), image.max() * 10,],
        )
    abscissa = np.array(np.mgrid[0 : image.shape[0] : 1, 0 : image.shape[1] : 1])
    # Perform the fitting
    popt, pcov = curve_fit(
        bivariate_normal,
        abscissa,
        image.flatten(),
        bounds=bounds,
        sigma=image_e.flatten(),
        p0=p0,
        maxfev=2000 * (len(p0) + 1),
    )
    # Determine uncertainty from covarience matrix
    p_sigma = np.sqrt(np.diag(pcov))
    return unp.uarray(popt, p_sigma), 4, 2


def univariate_normal(data, mu, sigma, offset, factor):
    """
    Produce a univariate normal distribution.

    Args:
        data (:pyttr:`array_like`): Abscissa data.
        mu (:py:attr:`float`): Mean (horizontal).
        sigma (:py:attr:`float`): Variance (horizontal).
        offset (:py:attr:`float`): Offset from the 0 for the ordinate, this is the background level.
        factor (:py:attr:`float`): Multiplicative factor for area of normal distribution.

    Returns:
        :py:attr:`array_like`: Ordinate data for uniivariate normal distribution.
    """
    # Creation of the bivariate normal distribution
    normal = norm(loc=mu, scale=sigma)
    return offset + normal.pdf(data).flatten() * factor


def fit_gaussian_1d(image, image_e, p0=None, bounds=None, axis=0):
    """
    Fit a one-dimensional Gaussian function with some ordinate offset to an image with uncertainty. 
    This is achieved by averaging in a given ``axis`` before performing the fit. 
    Return the results, and index of the offset.

    Args:
        image (:py:attr:`array_like`): The data to fit the Gaussian to.
        image_e (:py:attr:`array_like`): The data uncertainty.
        p0 (:py:attr:`list`, optional): An initial guess at the parameters. Defaults to values based on the image.
        bounds (:py:attr:`list` of :py:attr:`tuple`, optional): Bounds for the fitting. Defaults to values based on the image.
        axis (:py:attr:`int`): The dimension in which to perform the averaging.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: The results (with uncertainties) for each of the 6 parameters fit.
            - :py:attr:`int`: The index of the offset.
            - :py:attr:`None`: As it is not possible to describe the reflected peak width.
    """
    ordinate = image.mean(axis=axis)
    ordinate_e = image_e.mean(axis=axis)
    # Setting default values
    if p0 is None:
        p0 = [
            ordinate.shape[0] / 2,
            1,
            image.min(),
            image.max(),
        ]
    if bounds is None:
        bounds = (
            0,
            [ordinate.shape[0], 100, image.max(), image.max() * 10,],
        )
    # Perform the fitting
    popt, pcov = curve_fit(
        univariate_normal,
        np.arange(0, ordinate.shape[0], 1),
        ordinate,
        bounds=bounds,
        sigma=ordinate_e.flatten(),
        p0=p0,
        maxfev=2000 * (len(p0) + 1),
    )
    # Determine uncertainty from covarience matrix
    p_sigma = np.sqrt(np.diag(pcov))
    return unp.uarray(popt, p_sigma), 2, None 


"""
Often the detector is a lot larger than the reflected intensity peak. 
Therefore, we crop the image down, these functions help with this.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

import numpy as np


def crop_2d(array, array_e, x_start=0, x_end=-1, y_start=0, y_end=-1):
    """
    Crop the data (:py:attr:`array`) with some given start and stop point.

    Args:
        array (:py:attr:`array_like`): The intensity map collected by the 2 dimensional detector.
        array_e (:py:attr:`array_like`): Uncertainty map collected by the 2-D detector.
        x_start (:py:attr:`int`): Start point in x-axis.
        x_end (:py:attr:`int`): End point in x-axis.
        y_start (:py:attr:`int`): Start point in y-axis.
        y_end (:py:attr:`int`): End point in y-axis.

    Returns:
        :py:attr:`array_like`: A cropped intensity map.
    """
    cropped_array = array[y_start:y_end, x_start:x_end]
    if array_e is not None:
        cropped_error = array_e[y_start:y_end, x_start:x_end]
        return cropped_array, cropped_error
    return cropped_array


def crop_around_peak_2d(array, array_e, x_size=20, y_size=20):
    """
    Crop the data (`array`) around the most intense peak, creating an array
    of dimensions [x_size, y_size].

    Args:
        array (:py:attr:`array_like`): Intensity map collected by the 2 dimensional detector.
        array_e (:py:attr:`array_like`): Uncertainty map collected by the 2-D detector.
        x_size (:py:attr:`int`, optional): Size of the cropped image in x-axis. Defaults to :py:attr:`20`.
        y_size (:py:attr:`int`, optional): Size of the cropped image in y-axis. Defaults to :py:attr:`20`. 

    Returns:
        :py:attr:`array_like`: A cropped intensity map.
    """
    max_inten = np.unravel_index(np.argmax(array, axis=None), array.shape)
    half_x_size = int(x_size / 2)
    half_y_size = int(y_size / 2)
    return crop_2d(
            array,
            array_e,
            max_inten[1] - half_x_size,
            max_inten[1] + half_x_size,
            max_inten[0] - half_y_size,
            max_inten[0] + half_y_size,
        )



"""
As reflectometry measurements typicall consist of multiple scans at different attenutation, we must stitch these together.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey
import warnings
import numpy as np
from uncertainties import unumpy as unp
from scipy.stats import linregress

def correct_attentuation(scan_list):
    """
    Correct the attentuation level between a a series of elements in lists.

    Args:
        scans (:py:attr:`list` of :py:class:`islatu.refl_data.Scan`): Reflectometry scans.

    Returns:
        :py:attr:`list` of :py:class:`islatu.refl_data.Scan`: Reflectometry scans with attenuation corrected.
    """
    for i in range(len(scan_list) - 1):
        overlap_start = scan_list[i + 1].q[0].n
        overlap_end = scan_list[i].q[-1].n
        
        if overlap_start > overlap_end:
            warnings.warn('Using extrapolation to correct attenuation between scans {} and {}. Please double check these results.'.format(
                scan_list[i].file_path, scan_list[i+1].file_path))
            res = linregress(unp.nominal_values(scan_list[i].q[-2:]), np.log(unp.nominal_values(scan_list[i].R[-2:])))
            target_r = unp.exp(scan_list[i+1].q[0:2] * res.slope + res.intercept)
            vary_r = scan_list[i+1].R[0:2]
        else:
            overlap_start_index = np.argmin(np.abs(scan_list[i].q - overlap_start))
            overlap_end_index = np.argmin(np.abs(scan_list[i + 1].q - overlap_end))

            target_r = scan_list[i].R[overlap_start_index:]
            vary_r = scan_list[i + 1].R[: overlap_end_index + 1]

        ratio = target_r.mean() / vary_r.mean()
        
        scan_list[i + 1].R *= ratio
    return scan_list


def concatenate(scan_list):
    """
    Concatenate each of the datasets together.

    Args:
        scans (:py:attr:`list` of :py:class:`islatu.refl_data.Scan`): Reflectometry scans.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: q-values.
            - :py:attr:`array_like`: Reflected intensities.
    """
    reflected_intensity = np.array([])
    q_vectors = np.array([])
    for i in range(len(scan_list)):
        reflected_intensity = np.append(reflected_intensity, scan_list[i].R)
        q_vectors = np.append(q_vectors, scan_list[i].q)
    return q_vectors, reflected_intensity


def normalise_ter(q_vectors, reflected_intensity, max_q=0.1):
    """
    Find the total external reflection region and normalise this to 1.

    Args:
        max_q (:py:attr:`float`): The maximum q to be included in finding the critical angle.

    Returns:
        :py:attr:`array_like`: Reflected intensities.
    """
    q = unp.nominal_values(q_vectors)
    max_q_idx = q[np.where(q < max_q)].size
    if max_q_idx <= 1:
        end_of_ter_index = 1
    else:
        end_of_ter_index = np.argmax(
            np.abs(np.gradient(unp.nominal_values(reflected_intensity)[:max_q_idx]))
        )
    if end_of_ter_index == 0:
        end_of_ter_index = 1
    ter_region_mean_r = reflected_intensity[:end_of_ter_index].mean()
    reflected_intensity /= ter_region_mean_r
    return reflected_intensity


def rebin(q_vectors, reflected_intensity, new_q=None, number_of_q_vectors=400):
    """
    Rebin the data on a logarithmic q-scale.

    Args:
        new_q (:py:attr:`array_like`): Array of potential q-values. Defaults to :py:attr:`None`.
        number_of_q_vectors (:py:attr:`int`, optional): The max number of q-vectors to be using initially in the rebinning of the data. Defaults to :py:attr:`400`.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: q-values.
            - :py:attr:`array_like`: Reflected intensities.
    """
    if new_q is not None:
        new_q = new_q
    else:
        new_q = np.logspace(
            np.log10(q_vectors[0].n), np.log10(q_vectors[-1].n), number_of_q_vectors,
        )

    binned_q = unp.uarray(np.zeros_like(new_q), np.zeros_like(new_q))
    binned_r = unp.uarray(np.zeros_like(new_q), np.zeros_like(new_q))
    for i in range(len(new_q) - 1):
        count = 0
        for j, q in enumerate(q_vectors):
            if new_q[i] <= q < new_q[i + 1]:
                binned_q[i] += q
                binned_r[i] += reflected_intensity[j]
                count += 1
        if count > 0:
            binned_q[i] /= count
            binned_r[i] /= count
    cleaned_q = np.delete(binned_q, np.argwhere(binned_r == 0))
    cleaned_r = np.delete(binned_r, np.argwhere(binned_r == 0))

    return cleaned_q, cleaned_r



"""
The refl_data module includes two classes, :py:class:`~islatu.refl_data.Profile` and :py:class:`~islatu.refl_data.Scan` that are integral to the use of :py:mod`islatu` for the reduction of x-ray reflectometry data. 
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from os import path
import numpy as np
from scipy.constants import physical_constants
from scipy.interpolate import splev
from uncertainties import ufloat
from uncertainties import unumpy as unp
from islatu import corrections, image, stitching


class Profile:
    """
    This class stores information about the reflectometry profile.

    Attributes:
        scans (:py:attr:`list` of :py:class:`islatu.refl_data.Scan`): Reflectometry scans that make up the profile.
        q_vectors (:py:attr:`array_like`): q-vectors from the reflectometry scan, populated after concatentation.
        reflected_intensity (:py:attr:`array_like`): Reflected intensities from the reflectometry scan, populated after concatentation.
    
    Args:
        file_paths (:py:attr:`list`): List of files, one for each reflectometry scan.
        parser (:py:attr:`callable`): Parser function for the reflectometry scan files.
        q_axis_name (:py:attr:`str`, optional): Label for the q-axis in the scan. Defaults to :py:attr:`'q_axis_name'`. 
        theta_axis_name (:py:attr:`str`, optional): Label for the theta axis in the scan. Defaults to :py:attr:`'dcdtheta'`.
    """

    def __init__(
        self, file_paths, parser, q_axis_name="qdcd", theta_axis_name="dcdtheta", pixel_max=1000000, hot_pixel_threshold=200000, transpose=False
    ):
        self.scans = []
        try:
            kk = len(pixel_max)
        except TypeError:
            pixel_max = [pixel_max] * len(file_paths)
        try:
            jj = len(hot_pixel_threshold)
        except TypeError:
            hot_pixel_threshold = [hot_pixel_threshold] * len(file_paths)
        for i, f in enumerate(file_paths):
            self.scans.append(Scan(f, parser, q_axis_name, theta_axis_name, pixel_max=pixel_max[i], hot_pixel_threshold=hot_pixel_threshold[i], transpose=transpose))
        self.q_vectors = None
        self.reflected_intensity = None

    @property
    def R(self):
        """
        Reflected intensity values.

        Returns:
            :py:attr:`array_like`: Intensity values.
        """
        return unp.nominal_values(self.reflected_intensity)

    @property
    def dR(self):
        """
        Reflected intensity uncertainties.

        Returns:
            :py:attr:`array_like`: Intensity uncertainties.
        """
        return unp.std_devs(self.reflected_intensity)

    @property
    def q(self):
        """
        q-value values.

        Returns:
            :py:attr:`array_like`: q-value values.
        """
        return unp.nominal_values(self.q_vectors)

    @property
    def dq(self):
        """
        q-value uncertainties.

        Returns:
            :py:attr:`array_like`: q-value uncertainties.
        """
        return unp.std_devs(self.q_vectors)

    def crop_and_bkg_sub(
        self,
        crop_function,
        bkg_sub_function,
        crop_kwargs=None,
        bkg_sub_kwargs=None,
        progress=True,
    ):
        """
        Class method for the :func:`~islatu.refl_data.Scan.crop_and_bkg_sub` method for each :py:class:`~Scan` in the list.

        Args:
            crop_function (:py:attr:`callable`): Cropping function to be used.
            bkg_sub_function (:py:attr:`callable`): Background subtraction function to be used. 
            crop_kwargs (:py:attr:`dict`, optional): Keyword arguments for the cropping function. Defaults to :py:attr:`None`.
            bkg_sub_kwargs (:py:attr:`dict`, optional): Keyword arguments for the background subtraction function. Defaults to :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar. Requires the :py:mod:`tqdm` package. Defaults to :py:attr:`True`.
        """
        for s in self.scans:
            s.crop_and_bkg_sub(
                crop_function, bkg_sub_function, crop_kwargs, bkg_sub_kwargs, progress
            )

    def footprint_correction(self, beam_width, sample_size):
        """
        Class method for :func:`~islatu.refl_data.Scan.footprint_correction` for each :py:class:`~Scan` in the list.

        Args:
            beam_width (:py:attr:`float`): Width of incident beam, in metres.
            sample_size (:py:class:`uncertainties.core.Variable`): Width of sample in the dimension of the beam, in metres.
            theta (:py:attr:`float`): Incident angle, in degrees.
        """
        for s in self.scans:
            s.footprint_correction(beam_width, sample_size)

    def transmission_normalisation(self):
        """
        Perform the transmission correction (:func:`~islatu.refl_data.Scan.transmission_normalisation`) for each :py:class:`~Scan` in the list and fine attenutation correction (:func:`~islatu.stitching.correction_attenutation`). 
        """
        for s in self.scans:
            s.transmission_normalisation()
        self.scans = stitching.correct_attentuation(self.scans)

    def q_uncertainty_from_pixel(
        self,
        number_of_pixels=None,
        detector_distance=None,
        energy=None,
        pixel_size=172e-6,
    ):
        """
        Class method for :func:`~islatu.refl_data.Scan.q_uncertainty_from_pixel` for each :py:class:`~Scan` in the list.

       Args:
            number_of_pixels (:py:attr:`float`): Number of pixels of q uncertainty.
            detector_distance (:py:attr:`float`): Sample detector distance in metres
            energy (:py:attr:`float`): X-ray energy in keV
            pixel_size (:py:attr:`float`, optional): Pixel size in metres
        """
        for s in self.scans:
            s.q_uncertainty_from_pixel(
                number_of_pixels, detector_distance, energy, pixel_size
            )

    def qdcd_normalisation(self, itp):
        """
        Class method for :func:`~islatu.refl_data.Scan.qdcd_normalisation` for each :py:class:`~Scan` in the list.

        Args:
            normalisation_file (:py:attr:`str`): The ``.dat`` file that contains the normalisation data.
        """
        for s in self.scans:
            s.qdcd_normalisation(itp)

    def concatenate(self):
        """
        Class method for :func:`~islatu.stitching.concatenate`. 
        """
        self.q_vectors, self.reflected_intensity = stitching.concatenate(self.scans)

    def normalise_ter(self, max_q=0.1):
        """
        Class method for :func:`~islatu.stitching.normalise_ter`.

        Args:
            max_q (:py:attr:`float`): The maximum q to be included in finding the critical angle.
        """
        self.reflected_intensity = stitching.normalise_ter(
            self.q_vectors, self.reflected_intensity, max_q
        )

    def rebin(self, new_q=None, number_of_q_vectors=400):
        """
        Class method for :func:`islatu.stitching.rebin`. 

        Args:
            new_q (:py:attr:`array_like`): Array of potential q-values. Defaults to :py:attr:`None`.
            number_of_q_vectors (:py:attr:`int`, optional): The max number of q-vectors to be using initially in the rebinning of the data. Defaults to :py:attr:`400`.
        """
        self.q_vectors, self.reflected_intensity = stitching.rebin(
            self.q_vectors, self.reflected_intensity, new_q, number_of_q_vectors
        )


class Scan:
    """
    This class stores scan specific information.

    Attributes:
        file_paths (:py:attr:`str`): File path for scan file.
        metadata (:py:attr:`dict`): The metadata from the ``.dat`` file.
        data (:py:class:`pandas.DataFrame`): The data from the ``.dat`` file.
        q (:py:attr:`array_like`): The measured q-values in the scan.
        theta (:py:attr:`array_like`): The measured theta values in the scan.
        R (:py:attr:`array_like`): The reflected intensity values in the scan.
        n_pixels (:py:attr:`array_like`): The width in pixels of the beam in each image.
    
    Args:
        file_paths (:py:attr:`list`): List of files, one for each reflectometry scan.
        parser (:py:attr:`callable`): Parser function for the reflectometry scan files.
        q_axis_name (:py:attr:`str`, optional): Label for the q-axis in the scan. Defaults to :py:attr:`'q_axis_name'`. 
        theta_axis_name (:py:attr:`str`, optional): Label for the theta axis in the scan. Defaults to :py:attr:`'dcdtheta'`.
    """

    def __init__(
        self,
        file_path,
        parser,
        q_axis_name="qdcd",
        theta_axis_name="dcdtheta",
        energy=None,
        pixel_max=1000000,
        hot_pixel_threshold=200000,
        transpose=False,
    ):
        self.file_path = file_path
        self.metadata, self.data = parser(self.file_path)
        if q_axis_name is None:
            h = physical_constants["Planck constant in eV s"][0] * 1e-3
            c = physical_constants["speed of light in vacuum"][0] * 1e10
            if energy is None:
                energy = self.metadata["dcm1energy"][0]
            q = (
                energy
                * 4
                * np.pi
                * unp.sin(unp.radians(self.data[theta_axis_name]))
                / (h * c)
            )
            self.q = unp.uarray(q, np.zeros(self.data[theta_axis_name].size))
        else:
            self.q = unp.uarray(
                self.data[q_axis_name], np.zeros(self.data[q_axis_name].size)
            )
        self.data = self._check_files_exist()
        self.theta = unp.uarray(
            self.data[theta_axis_name], np.zeros(self.data[theta_axis_name].size),
        )
        self.R = unp.uarray(np.zeros(self.q.size), np.zeros(self.q.size),)
        self.n_pixels = np.zeros(self.q.size)
        self.pixel_max = pixel_max
        self.hot_pixel_threshold = hot_pixel_threshold
        self.transpose = transpose

    def __str__(self):
        """
        Custom string output.

        Returns:
            :py:attr:`str`: Description of scan.
        """
        return "The file: {} contains {} images from q = {:.4f} to q = {:.4f}.".format(
            self.file_path, self.q.size, self.q[0].n, self.q[-1].n
        )

    def __repr__(self):
        """
        Custom repr output.

        Returns:
            :py:attr:`str`: Description of scan.
        """
        return self.__str__()

    def _check_files_exist(self):
        """
        Check that image files exist.

        Returns: 
            :py:class:`pandas.DataFrame`: Modified data with corrected file paths.
        """
        iterator = _get_iterator(unp.nominal_values(self.q), False)
        for i in iterator:
            im_file = self.data["file"][i]
            if path.isfile(im_file):
                continue
            else:
                im_file = self.data["file"][i].split(path.sep)[-2:]
                im_file = path.join(im_file[0], im_file[1])
                im_file = path.join(path.dirname(self.file_path), im_file)
                if path.isfile(im_file):
                    self.data.iloc[i, self.data.keys().get_loc("file")] = im_file
                    continue
                else:
                    im_file = self.data["file"][i].split(path.sep)[-1]
                    im_file = path.join(path.dirname(self.file_path), im_file)
                    if path.isfile(im_file):
                        self.data.iloc[i, self.data.keys().get_loc("file")] = im_file
                        continue
                    else:
                        raise FileNotFoundError(
                            "The following image file could not be found: {}.".format(
                                self.data["file"][i]
                            )
                        )
        return self.data

    def crop_and_bkg_sub(
        self,
        crop_function,
        bkg_sub_function,
        crop_kwargs=None,
        bkg_sub_kwargs=None,
        progress=True,
    ):
        """
        Preform cropping and background subtraction on each image in the Scan.

        Args:
            crop_function (:py:attr:`callable`): Cropping function to be used.
            bkg_sub_function (:py:attr:`callable`): Background subtraction function to be used. 
            crop_kwargs (:py:attr:`dict`, optional): Keyword arguments for the cropping function. Defaults to :py:attr:`None`.
            bkg_sub_kwargs (:py:attr:`dict`, optional): Keyword arguments for the background subtraction function. Defaults to :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar. Requires the :py:mod:`tqdm` package. Defaults to :py:attr:`True`.
        """
        iterator = _get_iterator(unp.nominal_values(self.q), progress)
        to_remove = []
        for i in iterator:
            try:
                if self.data['roi1_maxval'][i] > self.pixel_max:
                    to_remove.append(i)
                    continue
                im = image.Image(self.data["file"][i], self.data, self.metadata, hot_pixel_threshold=self.hot_pixel_threshold, transpose=self.transpose)
                if crop_kwargs is None:
                    im.crop(crop_function)
                else:
                    im.crop(crop_function, **crop_kwargs)
                if bkg_sub_kwargs is None:
                    im.background_subtraction(bkg_sub_function)
                else:
                    im.background_subtraction(bkg_sub_function, **bkg_sub_kwargs)
                self.R[i] = ufloat(im.sum().n, im.sum().s)
                if im.n_pixel is None:
                    self.n_pixels[i] = None
                else:
                    self.n_pixels[i] = im.n_pixel.n
            except (ValueError, RuntimeError) as e:
                print(e)
                print('Current file: ', self.data["file"][i])
                to_remove.append(i)
                continue
        self.R = np.delete(self.R, to_remove)
        self.theta = np.delete(self.theta, to_remove)
        self.n_pixels = np.delete(self.n_pixels, to_remove)
        self.q = np.delete(self.q, to_remove)

    def footprint_correction(self, beam_width, sample_size):
        """
        Class method for :func:`islatu.corrections.footprint_correction`. 

        Args:
            beam_width (:py:attr:`float`): Width of incident beam, in metres.
            sample_size (:py:class:`uncertainties.core.Variable`): Width of sample in the dimension of the beam, in metres.
            theta (:py:attr:`float`): Incident angle, in degrees.
        """
        self.R /= corrections.footprint_correction(beam_width, sample_size, self.theta)

    def transmission_normalisation(self):
        """
        Perform the transmission correction. 
        """
        self.R /= float(self.metadata["transmission"][0])

    def q_uncertainty_from_pixel(
        self,
        number_of_pixels=None,
        detector_distance=None,
        energy=None,
        pixel_size=172e-6,
    ):
        """
        Calculate a q uncertainty from the area detector.

        Args:
            number_of_pixels (:py:attr:`float`): Number of pixels of q uncertainty.
            detector_distance (:py:attr:`float`): Sample detector distance in metres
            energy (:py:attr:`float`): X-ray energy in keV
            pixel_size (:py:attr:`float`, optional): Pixel size in metres

        Returns:
            :py:attr:`float`: Resulting q-uncertainty.
        """
        if number_of_pixels is None:
            number_of_pixels = self.n_pixels
        if detector_distance is None:
            detector_distance = self.metadata["diff1detdist"][0] * 1e-3
        if energy is None:
            energy = self.metadata["dcm1energy"][0]
        offset = np.arctan(
            pixel_size * 1.96 * number_of_pixels * 0.5 / (detector_distance)
        )
        h = physical_constants["Planck constant in eV s"][0] * 1e-3
        c = physical_constants["speed of light in vacuum"][0] * 1e10
        q_uncertainty = energy * 4 * np.pi * unp.sin(offset) / (h * c)
        self.q = unp.uarray(
            unp.nominal_values(self.q), unp.std_devs(self.q) + q_uncertainty
        )

    def qdcd_normalisation(self, itp):
        """
        Perform normalisation by DCD variance.

        Args:
            itp (:py:attr:`tuple`): Containing interpolation knots (:py:attr:`array_like`), B-spline coefficients (:py:attr:`array_like`), and degree of spline (:py:attr:`int`).
        """
        self.R /= splev(unp.nominal_values(self.q), itp)


def _get_iterator(q, progress):
    """
    Create a q-value iterator.

    Args:
        q (:py:attr:`array_like`): q-values.
        progress (:py:attr:`bool`): Show progress bar.
    
    Returns:
        :py:attr:`range` or :py:class:`tqdm.std.tqdm`: Iterator object. 
    """
    iterator = range(len(q))
    if progress:
        try:
            from tqdm import tqdm

            iterator = tqdm(range(len(q)))
        except ModuleNotFoundError:
            print(
                "For the progress bar, you need to have the tqdm package "
                "installed. No progress bar will be shown"
            )
    return iterator



"""
Reflectometry data must be corrected as a part of reduction. 
These functions facilitate this, including the footprint and DCD q-variance corrections.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

import numpy as np
from scipy.stats import norm
from uncertainties import unumpy as unp
from islatu.io import i07_dat_parser
from scipy.interpolate import splrep


def footprint_correction(beam_width, sample_size, theta):
    """
    The factor by which the intensity should be multiplied to account for the
    scattering geometry, where the beam is Gaussian in shape.

    Args:
        beam_width (:py:attr:`float`): Width of incident beam, in metres.
        sample_size (:py:class:`uncertainties.core.Variable`): Width of sample in the dimension of the beam, in metres.
        theta (:py:attr:`float`): Incident angle, in degrees.

    Returns:
        (:py:class:`uncertainties.core.Variable`): Correction factor.
    """
    beam_sd = beam_width / 2 / np.sqrt(2 * np.log(2))
    length = sample_size * unp.sin(unp.radians(theta))
    mid = unp.nominal_values(length) / 2.0 / beam_sd
    upper = (unp.nominal_values(length) + unp.std_devs(length)) / 2.0 / beam_sd
    lower = (unp.nominal_values(length) - unp.std_devs(length)) / 2.0 / beam_sd
    probability = 2.0 * (
        unp.uarray(norm.cdf(mid), (norm.cdf(upper) - norm.cdf(lower)) / 2) - 0.5
    )
    return probability


def get_interpolator(
    file_path, parser, q_axis_name="qdcd_", intensity_axis_name="adc2"
):
    """
    Get an interpolator object from scipy, this is useful for the DCD q-normalisation step.

    Args:
        file_path (:py:attr:`str`): File path to the normalisation file.
        parser (:py:attr:`callable`): Parser function for the normalisation file.
        q_axis_name (:py:attr:`str`, optional): Label for the q-value in the normalisation file. Defaults to ``'qdcd_'``.
        intensity_axis_name (:py:attr:`str`, optional): Label for the intensity in the normalisation file. Defaults to ``'adc2'``.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: Interpolation knots.
            - :py:attr:`array_like`: B-spline coefficients.
            - :py:attr:`int`: Degree of spline.
    """
    normalisation_data = parser(file_path)[1]
    return splrep(
        normalisation_data[q_axis_name], normalisation_data[intensity_axis_name]
    )



"""
The two-dimension detector generates images of the reflected intensity. 
The purpose of this class is the investigation and manipulation of these images.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

import os
import numpy as np
from matplotlib.pyplot import imshow
from PIL import Image as PILIm
from uncertainties import unumpy as unp


class Image:
    """
    This class stores information about the detector images.

    Attributes:
        file_path (:py:attr:`str`): File path for the image.
        data (:py:class:`pandas.DataFrame`): Experimental data about the measurement.
        metadata (:py:attr:`dict`): Metadata regarding the measurement.
        array (:py:attr:`array_like`): The image described as an array.
        bkg (:py:class:`uncertainties.cores.Variable`): The background that was subtracted from the image.
        n_pixels (:py:attr:`float`): The width of the peak in number of pixels, used to calculate an uncertainty in q on the detector.

    Args:
        file_path (:py:attr:`str`): The file path for the image.
        data (:py:class:`pandas.DataFrame`, optional): Experimental data about the measurement. Defaults to :py:attr:`None`.
        metadata (:py:attr:`dict`, optional): Metadata regarding the measurement. Defaults to :py:attr:`None`.
        transpose (:py:attr:`bool`, optional): Should the data be rotated by 90 degrees? Defaults to :py:attr:`False`.
        hot_pixel_threshold (:py:attr:`int`, optional): The number of counts above which a pixel should be assessed to determine if it is hot. Defaults to :py:attr:`200000`.
        pixel_maximum (:py:attr:`int`, optional): The number of counts above which a pixel should be assessed to determine if it is hot. Defaults to :py:attr:`500000`.
        pixel_minimum (:py:attr:`int`, optional): The number of counts above which a pixel should be assessed to determine if it is hot. Defaults to :py:attr:`0`.
    """

    def __init__(
        self,
        file_path,
        data=None,
        metadata=None,
        transpose=False,
        hot_pixel_threshold=200000,
        pixel_minimum=0
    ):
        """
        Initialisation of the :py:class:`islatu.image.Image` class, includes running hot pixel check and assigning uncertainties.
        """
        self.file_path = file_path
        self.data = data
        self.metadata = metadata
        img = PILIm.open(file_path)
        array = np.array(img)
        img.close()
        if transpose:
            array = array.T
        # Remove hot pixels
        array = _find_hot_pixels(array, threshold=hot_pixel_threshold)
        array[np.where(array < pixel_minimum)] = 0
        array_error = np.sqrt(array)
        array_error[np.where(array == 0)] = 1
        self.array = unp.uarray(array, array_error)
        self.bkg = 0
        self.n_pixel = 0

    @property
    def nominal_values(self):
        """
        Get the nominal values of the image array.

        Returns:
            :py:attr:`array_like`: Nominal values of image.
        """
        return unp.nominal_values(self.array)

    @property
    def std_devs(self):
        """
        Get the standard deviation values of the image array.

        Returns:
            :py:attr:`array_like`: Standard deviation values of image.
        """
        return unp.std_devs(self.array)

    @property
    def n(self):
        """
        Get the nominal values of the image array.

        Returns:
            :py:attr:`array_like`: Nominal values of image.
        """
        return unp.nominal_values(self.array)

    @property
    def s(self):
        """
        Get the standard deviation values of the image array.

        Returns:
            :py:attr:`array_like`: Standard deviation values of image.
        """
        return unp.std_devs(self.array)

    @property
    def shape(self):
        """
        Array shape

        Returns:
            :py:attr:`tuple` of :py:attr:`int`: The shape of the image.
        """
        return unp.nominal_values(self.array).shape

    def show(self):
        """
        Show the image.

        Return:
            :py:class:`matplotlib.figure.Figure`: Matplotlib imshow of array.
        """
        return im.show(self.n)

    def __repr__(self):
        """
        Custom representation.

        Returns:
            :py:attr:`array_like`: Image array.
        """
        return self.array

    def __str__(self):
        """
        Custom string.

        Returns:
            :py:attr:`array_like`: Image array.
        """
        return self.array

    def crop(self, crop_function, **kwargs):
        """
        Perform an image crop based on some function.

        Args:
            crop_function (:py:attr:`callable`): The function to crop the data.
            **kwargs (:py:attr:`dict`): The crop function keyword arguments.
        """
        self.array = unp.uarray(*crop_function(self.n, self.s, **kwargs))

    def background_subtraction(self, background_subtraction_function, **kwargs):
        """
        Perform a background subtraction based on some function.

        Args:
            background_subtraction_function (:py:attr:`callable`): The function to model the data and therefore remove the background.
            **kwargs (:py:attr:`dict`): The background substraction function keyword arguments.
        """
        bkg_popt, bkg_idx, pixel_idx = background_subtraction_function(
            self.n, self.s, **kwargs
        )
        self.bkg = bkg_popt[bkg_idx]
        if pixel_idx is None:
            self.n_pixel = None
        else:
            self.n_pixel = bkg_popt[pixel_idx]
        self.array -= self.bkg

    def sum(self, axis=None):
        """
        Perform a summation on the image

        Args:
            axis (:py:attr:`int`, optional): The axis of the array to perform the summation over. Defaults to :py:attr:`None`.
        """
        return self.array.sum(axis)


def _find_hot_pixels(array, threshold=200000):
    """
    Find some dead pixels and assign them with some local average value.

    Args:
        array (:py:attr:`array_like`): NumPy array describing the image.
        threshold (:py:attr:`int`, optional): The number of counts above which a pixel should be assessed to determine if it is hot. Defaults to :py:attr:`200000`.

    Returns:
        :py:attr:`array_like`: NumPy array where hot pixels have been removed.
    """
    sorted_array = np.sort(array.flatten())[::-1]
    for i in sorted_array:
        if i >= threshold:
            pos_a, pos_b = np.where(array == i)
            lower_a = pos_a[0] - 1
            upper_a = pos_a[0] + 2
            lower_b = pos_b[0] - 1
            upper_b = pos_b[0] + 2
            if pos_a[0] == 0:
                lower_a = 0
            elif pos_a[0] == array.shape[0] - 1:
                upper_a = array.shape[0]
            if pos_b[0] == 0:
                lower_b = 0
            elif pos_b[0] == array.shape[1] - 1:
                upper_b = array.shape[1]
            local = np.copy(array[lower_a:upper_a, lower_b:upper_b])
            local[np.where(local == i)] = 0
            local_average = np.mean(local)
            local_std = np.std(local)
            if i > local_average + 2 * local_std:
                array[pos_a[0], pos_b[0]] = local_average
        else:
            break
    return array



#! /usr/bin/env python
"""
setup.py for islatu

@author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)
"""

# System imports
import io
from os import path
from setuptools import setup, find_packages

PACKAGES = find_packages()

# versioning
MAJOR = 0
MINOR = 0
MICRO = 24 
ISRELEASED = False
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


THIS_DIRECTORY = path.abspath(path.dirname(__file__))
with io.open(path.join(THIS_DIRECTORY, 'README.md')) as f:
    LONG_DESCRIPTION = f.read()

INFO = {
        'name': 'islatu',
        'description': 'X-ray reflectometry reduction in Python.',
        'author': 'Andrew R. McCluskey',
        'author_email': 'andrew.mccluskey@diamond.ac.uk ',
        'packages': PACKAGES,
        'include_package_data': True,
        'setup_requires': ['cython', 'numpy', 'scipy', 'uncertainties', 'pillow', 'pandas'],
        'install_requires': ['cython', 'numpy', 'scipy', 'uncertainties', 'pillow', 'pandas'],
        'version': VERSION,
        'license': 'MIT',
        'long_description': LONG_DESCRIPTION,
        'long_description_content_type': 'text/markdown',
        'classifiers': ['Development Status :: 2 - Pre-Alpha',
                        'Intended Audience :: Science/Research',
                        'License :: OSI Approved :: MIT License',
                        'Natural Language :: English',
                        'Operating System :: OS Independent',
                        'Programming Language :: Python :: 3.5',
                        'Programming Language :: Python :: 3.6',
                        'Programming Language :: Python :: 3.7',
                        'Programming Language :: Python :: 3.8',
                        'Programming Language :: Python :: 3.9',
                        'Topic :: Scientific/Engineering',
                        'Topic :: Scientific/Engineering :: Chemistry',
                        'Topic :: Scientific/Engineering :: Physics']
        }

####################################################################
# this is where setup starts
####################################################################


def setup_package():
    """
    Runs package setup
    """
    setup(**INFO)


if __name__ == '__main__':
    setup_package()
